"""This module is for the `For` loop class implementation.
"""

from typing import Any
from typing import Dict
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._expression.indent_num import Indent
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.array import Array
from apysc._type.dictionary import Dictionary
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String

T = TypeVar('T', Int, String, Number)


class For(Generic[T]):
    """
    A class to append for the (loop) expression.

    References
    ----------
    - For document
        - https://simon-ritchie.github.io/apysc/for.html

    Examples
    --------
    >>> import apysc as ap
    >>> arr: ap.Array = ap.Array(range(3))
    >>> with ap.For(arr) as i:
    ...     ap.trace('Loop index is:', i)
    """

    _arr_or_dict: Union[Array, Dictionary]
    _locals: Dict[str, Any]
    _globals: Dict[str, Any]
    _snapshot_name: str
    _indent: Indent

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='For')
    def __init__(
            self, arr_or_dict: Union[Array, Dictionary],
            *,
            locals_: Optional[Dict[str, Any]] = None,
            globals_: Optional[Dict[str, Any]] = None) -> None:
        """
        A class to append for the (loop) expression.

        Parameters
        ----------
        arr_or_dict : Array or Dictionary
            Array or Dictionary instance to iterate.
        locals_ : dict or None, default None
            Current scope's local variables. Set locals()
            value to this argument. If specified, this interface
            reverts all local scope VariableNameInterface
            variables (like Int, Sprite) at the end of a `For`
            scope. This setting is useful when you don't want to
            update each variable by implementing the `For` scope.
        globals_ : dict or None, default None
            Current scope's global variables. Set globals()
            value to this argument. This setting works
            the same way as the locals_ argument.

        References
        ----------
        - For document
            - https://simon-ritchie.github.io/apysc/for.html

        Examples
        --------
        >>> import apysc as ap
        >>> arr: ap.Array = ap.Array(range(3))
        >>> with ap.For(arr) as i:
        ...     ap.trace('Loop index is:', i)
        """
        self._validate_arr_or_dict_val_type(arr_or_dict=arr_or_dict)
        if locals_ is None:
            locals_ = {}
        if globals_ is None:
            globals_ = {}
        self._arr_or_dict = arr_or_dict
        self._locals = locals_
        self._globals = globals_
        self._indent = Indent()

    def _validate_arr_or_dict_val_type(
            self, *, arr_or_dict: Union[Array, Dictionary]) -> None:
        """
        Validate loop value type is Array of Dictionary.

        Parameters
        ----------
        arr_or_dict : Array or Dictionary
            Value to be checked.

        Raises
        ------
        TypeError
            If a value type is neither Array nor Dictionary.
        """
        if isinstance(arr_or_dict, (Array, Dictionary)):
            return
        raise TypeError(
            'Specified value type is neither Array nor Dictionary: '
            f'{type(arr_or_dict)}')

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='For')
    def __enter__(self) -> T:
        """
        This class calls this method at the with-statement.

        Returns
        -------
        i_or_key : Int or String
            Loop index or dictionary key.
        """
        import apysc as ap
        from apysc._loop import loop_count
        from apysc._type import revert_interface
        loop_count.increment_current_loop_count()
        self._snapshot_name = \
            revert_interface.make_snapshots_of_each_scope_vars(
                locals_=self._locals, globals_=self._globals)
        i_or_key: Union[ap.Int, ap.String]
        if isinstance(self._arr_or_dict, ap.Array):
            i_or_key = ap.Int(0)
            self._append_arr_enter_expression(i=i_or_key)
        elif isinstance(self._arr_or_dict, ap.Dictionary):
            if self._arr_or_dict._value:
                key: str = str(list(self._arr_or_dict._value.keys())[0])
            else:
                key = ''
            i_or_key = ap.String(key)
            self._append_dict_enter_expression(key=i_or_key)
        self._indent.__enter__()
        return i_or_key  # type: ignore

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='For')
    def __exit__(self, *args: Any) -> None:
        """
        This class calls this method at the with-statement.
        """
        import apysc as ap
        from apysc._expression import last_scope
        from apysc._expression.last_scope import LastScope
        from apysc._loop import loop_count
        from apysc._type import revert_interface
        loop_count.decrement_current_loop_count()
        revert_interface.revert_each_scope_vars(
            snapshot_name=self._snapshot_name,
            locals_=self._locals, globals_=self._globals)
        self._indent.__exit__()
        ap.append_js_expression(expression='}')
        last_scope.set_last_scope(value=LastScope.FOR)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='For')
    def _append_arr_enter_expression(self, *, i: Int) -> None:
        """
        Append for loop start expression (for Array value).

        Parameters
        ----------
        i : Int
            Loop index value.
        """
        import apysc as ap
        i_name: str = i.variable_name
        expression: str = (
            f'var length = {self._arr_or_dict.variable_name}.length;\n'
            f'for ({i_name} = 0; {i_name} < length; {i_name}++) {{'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='For')
    def _append_dict_enter_expression(self, *, key: String) -> None:
        """
        Append for loop start expression (for Dictionary value).

        Parameters
        ----------
        key : String
            Loop (dictionary) key value.
        """
        import apysc as ap
        key_name: str = key.variable_name
        expression: str = (
            f'for (var {key_name} in '
            f'{self._arr_or_dict.variable_name}) {{'
        )
        ap.append_js_expression(expression=expression)



"""This module is for the translation mapping data of the
following document:

Document file: assertion_basic_behavior.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# JavaScript assertion interface basic behaviors':
    '',

    'This page explains the JavaScript assertion interface basic behavior.':
    '',

    '## Interface names':
    '',

    'Each JavaScript assertion interface has the prefix of the `assert_` (e.g., `assert_equal`, `assert_true`, and so on).\n\nThese interfaces are positioned in the root package so you can use them, for example, `ap.assert_equal(...)`.':  # noqa
    '',

    '## Assertion results':
    '',

    'These interfaces display the results on the browser console, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=10, right=int_1)\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_results/\')\n```':  # noqa
    '',

    'This code displays the information message on the browser console, like this (please press the F12 key to see):':  # noqa
    '',

    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 10 right value: 10\n```':  # noqa
    '',

    '<iframe src="static/assertion_basic_behavior_results/index.html" width="0" height="0"></iframe>\n\nIf the assertion fails, then an error message also is displayed on the browser console:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=11, right=int_1)\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_results_failed/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 11 right value: 10\n...\nAssertion failed:\n...\n```':  # noqa
    '',

    '<iframe src="static/assertion_basic_behavior_results_failed/index.html" width="0" height="0"></iframe>':  # noqa
    '',

    '## Optional msg argument':
    '',

    'Each assertion interface has the `msg` optional argument. If you provide this argument, the error message is displayed on the browser console when an assertion fails.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstage: ap.Stage = ap.Stage(\n    stage_width=0, stage_height=0, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nint_1: ap.Int = ap.Int(10)\nap.assert_equal(left=11, right=int_1, msg=\'Values are not equal!\')\nap.save_overall_html(\n    dest_dir_path=\'assertion_basic_behavior_msg/\')\n```':  # noqa
    '',

    '':
    '',

    '```\n[assert_equal]\nRight-side variable name: i_11\nLeft value: 11 right value: 10\n...\nAssertion failed: Values are not equal!\n```':  # noqa
    '',

    '<iframe src="static/assertion_basic_behavior_msg/index.html" width="0" height="0"></iframe>':  # noqa
    '',

}



"""Base class implementation for the animation.
"""

from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union

from apysc._animation.easing import Easing
from apysc._event import animation_event
from apysc._event.custom_event_interface import CustomEventInterface
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.variable_name_interface import VariableNameInterface

_T = TypeVar('_T', bound=VariableNameInterface)
_O = TypeVar('_O')
_Handler = Callable[['animation_event.AnimationEvent', _O], None]


class AnimationBase(
        VariableNameInterface, CustomEventInterface, Generic[_T], ABC):

    _target: _T
    _duration: Int
    _delay: Int
    _easing: Easing
    _started: Boolean

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='AnimationBase')
    def __init__(self, *, variable_name: str) -> None:
        """
        Base class for each animation setting.

        Parameters
        ----------
        variable_name : str
            Variable name.
        """
        self.variable_name = variable_name
        self._started = Boolean(False)

    @abstractmethod
    def _get_animation_func_expression(self) -> str:
        """
        Get an animation function expression.
        """

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='AnimationBase')
    def _set_basic_animation_settings(
            self,
            *,
            target: _T,
            duration: Union[int, Int] = 3000,
            delay: Union[int, Int] = 0,
            easing: Easing = Easing.LINEAR) -> None:
        """
        Set the basic animation settings.

        Parameters
        ----------
        target : VariableNameInterface
            A target instance of the animation target
            (e.g., `DisplayObject` instance).
        duration : int or Int, default 3000
            Milliseconds before an animation ends.
        delay : int or Int, default 0
            Milliseconds before an animation starts.
        easing : Easing, default Easing.LINEAR
            Easing setting. If None, Linear calculation is used.
        """
        from apysc._converter import to_apysc_val_from_builtin
        self._target = target
        self._duration = to_apysc_val_from_builtin.\
            get_copied_int_from_builtin_val(integer=duration)
        self._delay = to_apysc_val_from_builtin.\
            get_copied_int_from_builtin_val(integer=delay)
        self._easing = easing

    def _get_animation_basic_expression(self) -> str:
        """
        Get an animation basic expression string.

        Returns
        -------
        expression : str
            An animation basic expression string.
        """
        expression: str = (
            f'{self._target.variable_name}'
            '\n  .animate({'
            f'\n    duration: {self._duration.variable_name},'
            f'\n    delay: {self._delay.variable_name}}})'
            f'\n  .ease({self._easing.value})'
        )
        expression += self._get_animation_complete_handler_expression()
        return expression

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='AnimationBase')
    def start(self) -> 'AnimationBase':
        """
        Start an animation with current settings.

        Returns
        -------
        self : AnimatonBase
            This instance.

        References
        ----------
        - AnimationBase class start interface
            - https://simon-ritchie.github.io/apysc/animation_base_start.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> _ = rectangle.animation_x(x=100).start()
        """
        import apysc as ap
        expression: str = self._get_animation_basic_expression()
        animation_expresssion: str = self._get_animation_func_expression()
        expression += animation_expresssion
        ap.append_js_expression(expression=expression)
        self._started.value = True
        return self

    def _get_animation_complete_handler_expression(self) -> str:
        """
        Get an expression of the animation complete handlers setting.

        Returns
        -------
        expression : str
            Target expression string.
            e.g., '\n  .after(handler_name)'

        Notes
        -----
        If multiple handlers are registered, then this method
        returns multiple lines expression.
        """
        from apysc._event.custom_event_type import CustomEventType
        event_type: str = CustomEventType.ANIMATION_COMPLETE.value
        self._initialize_custom_event_handlers_if_not_initialized(
            custom_event_type_str=event_type)
        if event_type not in self._custom_event_handlers:
            return ''
        expression: str = ''
        for handler_name in self._custom_event_handlers[event_type].keys():
            expression += f'\n  .after({handler_name})'
        return expression

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='AnimationBase')
    def animation_complete(
            self, handler: _Handler[_O], *,
            options: Optional[_O] = None) -> 'AnimationBase':
        """
        Add an animation complete event listener setting.

        Notes
        -----
        This interface can only use before an animation starts

        Parameters
        ----------
        handler : _Handler
            A callable that an instance calls when an animation
            is complete.
        options : dict or None, default None
            Optional arguments dictionary to be passed to a handler.

        Returns
        -------
        self : AnimatonBase
            This instance.

        Raises
        ------
        Exception
            If calling this interface after an animation starts

        References
        ----------
        - AnimationBase class animation_complete interface document
            - https://simon-ritchie.github.io/apysc/animation_complete.html
        - About the handler options’ type document
            - https://simon-ritchie.github.io/apysc/about_handler_options_type.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> def on_animation_complete(
        ...         e: ap.AnimationEvent[ap.Rectangle],
        ...         options: dict) -> None:
        ...     ap.trace('Animation completed!')
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> _ = rectangle.animation_x(
        ...     x=100,
        ... ).animation_complete(on_animation_complete).start()
        """
        import apysc as ap
        from apysc._event.custom_event_type import CustomEventType
        from apysc._validation.handler_options_validation import \
            validate_options_type
        self._validate_animation_not_started()
        validate_options_type(options=options)
        e: ap.AnimationEvent[_T] = ap.AnimationEvent(this=self)
        in_handler_head_expression: str = \
            self._get_complete_event_in_handler_head_expression()
        self.bind_custom_event(
            custom_event_type=CustomEventType.ANIMATION_COMPLETE,
            handler=handler,
            e=e,
            options=options,
            in_handler_head_expression=in_handler_head_expression)
        return self

    @abstractmethod
    def _get_complete_event_in_handler_head_expression(self) -> str:
        """
        Get an expression to be inserted into the complete event
        handler's head.
        """

    def _validate_animation_not_started(self) -> None:
        """
        Validate whether an animation has not already started.

        Raises
        ------
        Exception
            If an animation has already started.
        """
        if not self._started._value:
            return
        raise Exception(
            'This interface can not be called after the animation '
            'is started.')

    @property
    def target(self) -> _T:
        """
        Get an animation target instance.

        Returns
        -------
        target : VariableNameInterface
            An animation target instance.

        References
        ----------
        - AnimationBase class target property interface document
            - https://simon-ritchie.github.io/apysc/animation_base_target.html

        Examples
        --------
        >>> import apysc as ap
        >>> def on_animation_complete(
        ...         e: ap.AnimationEvent[ap.Rectangle],
        ...         options: dict) -> None:
        ...     rectangle: ap.Rectangle = e.this.target
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> _ = rectangle.animation_x(
        ...     x=100,
        ... ).animation_complete(on_animation_complete).start()
        """
        return self._target



"""This module is for the translation mapping data of the
following document:

Document file: return.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Return':
    '',

    'This page explains the `Return` class.\n\nBefore reading on, maybe it is helpful to read the following page (the apysc uses the `Return` class for the same reason of each apysc data type):\n\n- [Why the apysc library doesn\'t use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '',

    '## What is the Return class?':
    '',

    'The `Return` class behaves to append the `return;` JavaScript code. Therefore, this class can be used only in an event handler (function or method) scope.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `Return` class constructor accepts no arguments. You can use this interface with the branch condition, for example, the `ap.If` class.\n\nThe following example changes the rectangle fill color when you click it. Each `ap.If` branch instantiate `Return` class, so the code applies the changing of fill color one by one:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(\n        e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : ap.MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this\n    fill_color: ap.String = rectangle.fill_color\n    with ap.If(fill_color == \'#00aaff\'):\n        rectangle.fill_color = ap.String(\'#ff00aa\')\n        ap.Return()\n    with ap.If(fill_color == \'#ff00aa\'):\n        rectangle.fill_color = ap.String(\'#00ffaa\')\n        ap.Return()\n    with ap.If(fill_color == \'#00ffaa\'):\n        rectangle.fill_color = ap.String(\'#00aaff\')\n        ap.Return()\n\n\nap.Stage(\n    stage_width=150, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#00aaff\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'return_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/return_basic_usage/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## Return API':
    '',

    '<!-- Docstring: apysc._type._return.Return.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self) -> None`<hr>\n\n**[Interface summary]** Class for the return expression.<hr>\n\n**[Notes]**\n\nThis class can be instantiated only in an event handler scope.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> def on_timer(e: ap.TimerEvent, options: dict) -> None:\n...     """\n...     The handler that the timer calls.\n...\n```':  # noqa
    '',

}



"""Definition of the FPS enum.
"""

from enum import Enum
from typing import Union


class FPSDefinition:

    _fps: int
    _milisecond_intervals: Union[int, float]

    def __init__(
            self, *, fps: int,
            milisecond_intervals: Union[int, float]) -> None:
        """
        FPS definition class.

        Parameters
        ----------
        fps : int
            FPS value, such as 30, 60.
        milisecond_intervals : int or float
            FPS value in milisecond intervals, such as 33.333...
        """
        self._fps = fps
        self._milisecond_intervals = milisecond_intervals

    @property
    def milisecond_intervals(self) -> float:
        """
        Get a FPS value in milisecond intervals.

        Returns
        -------
        milisecond_interval : float
            FPS value in milisecond intervals, such as 33.333...
        """
        return self._milisecond_intervals


class FPS(Enum):
    """
    Definition of the FPS enum.

    References
    ----------
    - FPS enum document
        - https://simon-ritchie.github.io/apysc/fps.html

    Examples
    --------
    >>> import apysc as ap
    >>> def on_timer(e: ap.TimerEvent, options: dict) -> None:
    ...     pass
    >>> ap.Timer(on_timer, delay=ap.FPS.FPS_60).start()
    """

    FPS_5 = FPSDefinition(fps=5, milisecond_intervals=200)
    FPS_10 = FPSDefinition(fps=10, milisecond_intervals=100)
    FPS_15 = FPSDefinition(fps=15, milisecond_intervals=66.6666667)
    FPS_20 = FPSDefinition(fps=20, milisecond_intervals=50)
    FPS_25 = FPSDefinition(fps=25, milisecond_intervals=40)
    FPS_30 = FPSDefinition(fps=30, milisecond_intervals=33.3333333)
    FPS_35 = FPSDefinition(fps=35, milisecond_intervals=28.5714286)
    FPS_40 = FPSDefinition(fps=40, milisecond_intervals=25)
    FPS_45 = FPSDefinition(fps=45, milisecond_intervals=22.2222222)
    FPS_50 = FPSDefinition(fps=50, milisecond_intervals=20)
    FPS_55 = FPSDefinition(fps=55, milisecond_intervals=18.1818182)
    FPS_60 = FPSDefinition(fps=60, milisecond_intervals=16.6666667)



"""Implementations for the current loop number
(the For or the other loop class nested number).
"""

from typing import Optional
from typing import Tuple


def get_current_loop_count() -> int:
    """
    Get the current loop count number.

    Returns
    -------
    loop_count : int
        Current loop count.
    """
    from apysc._expression import expression_data_util
    query: str = (
        'SELECT count FROM '
        f'{expression_data_util.TableName.LOOP_COUNT.value} '
        'LIMIT 1;'
    )
    expression_data_util.exec_query(sql=query)
    result: Optional[Tuple[int]] = expression_data_util.cursor.fetchone()
    if result is None:
        return 0
    return result[0]


def _save_loop_count(*, loop_count: int) -> None:
    """
    Save a loop count value.

    Parameters
    ----------
    loop_count : int
        A loop count value.
    """
    from apysc._expression import expression_data_util
    table_name: str = expression_data_util.TableName.LOOP_COUNT.value
    query: str = f'DELETE FROM {table_name};'
    expression_data_util.exec_query(sql=query, commit=False)
    query = f'INSERT INTO {table_name}(count) VALUES ({loop_count});'
    expression_data_util.exec_query(sql=query)


def increment_current_loop_count() -> None:
    """
    Increment the current loop count number.
    """
    current_loop_count: int = get_current_loop_count()
    _save_loop_count(loop_count=current_loop_count + 1)


def decrement_current_loop_count() -> None:
    """
    Decrement the current loop count number.
    """
    current_loop_count: int = get_current_loop_count()
    current_loop_count -= 1
    current_loop_count = max(0, current_loop_count)
    _save_loop_count(loop_count=current_loop_count)



"""Class implementation for graphics clear method related interface.
"""


from typing import Optional

from apysc._display.display_object import DisplayObject
from apysc._display.line_dash_dot_setting import LineDashDotSetting
from apysc._display.line_dash_setting import LineDashSetting
from apysc._display.line_dot_setting import LineDotSetting
from apysc._display.line_round_dot_setting import LineRoundDotSetting
from apysc._display.polyline import Polyline
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.array import Array
from apysc._type.int import Int
from apysc._type.number import Number
from apysc._type.string import String


class GraphicsClearInterface:

    _fill_color: String
    _fill_alpha: Number
    _line_color: String
    _line_thickness: Int
    _line_alpha: Number
    _children: Array[DisplayObject]
    _current_line: Optional[Polyline]
    _line_cap: String
    _line_joints: String
    _line_dot_setting: Optional[LineDotSetting]
    _line_dash_setting: Optional[LineDashSetting]
    _line_round_dot_setting: Optional[LineRoundDotSetting]
    _line_dash_dot_setting: Optional[LineDashDotSetting]

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='GraphicsClearInterface')
    def clear(self) -> None:
        """
        Clear all graphics and reset fill and line settings.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af')
        >>> _ = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> _ = sprite.graphics.draw_rect(
        ...     x=150, y=50, width=50, height=50)
        >>> sprite.graphics.num_children
        Int(2)

        >>> sprite.graphics.fill_color
        String('#00aaff')

        >>> sprite.graphics.clear()
        >>> sprite.graphics.num_children
        Int(0)

        >>> sprite.graphics.fill_color
        String('')
        """
        import apysc as ap
        from apysc._display.begin_fill_interface import BeginFillInterface
        from apysc._display.child_interface import ChildInterface
        from apysc._display.fill_alpha_interface import FillAlphaInterface
        from apysc._display.fill_color_interface import FillColorInterface
        from apysc._display.line_alpha_interface import LineAlphaInterface
        from apysc._display.line_color_interface import LineColorInterface
        from apysc._display.line_style_interface import LineStyleInterface
        from apysc._display.line_thickness_interface import \
            LineThicknessInterface
        if isinstance(self, (FillColorInterface, BeginFillInterface)):
            self._initialize_fill_color_if_not_initialized()
        self._fill_color.value = ''
        if isinstance(self, (FillAlphaInterface, BeginFillInterface)):
            self._initialize_fill_alpha_if_not_initialized()
        self._fill_alpha.value = 1.0
        if isinstance(self, (LineColorInterface, LineStyleInterface)):
            self._initialize_line_color_if_not_initialized()
        self._line_color.value = ''
        if isinstance(self, (LineThicknessInterface, LineStyleInterface)):
            self._initialize_line_thickness_if_not_initialized()
        self._line_thickness.value = 1
        if isinstance(self, (LineAlphaInterface, LineStyleInterface)):
            self._initialize_line_alpha_if_not_initialized()
        self._line_alpha.value = 1.0
        if isinstance(self, ChildInterface):
            self._initialize_children_if_not_initialized()
        while self._children:
            self._children[0].remove_from_parent()
        if hasattr(self, '_current_line'):
            self._current_line = None
        if hasattr(self, '_line_cap'):
            self._line_cap = ap.String(ap.LineCaps.BUTT.value)
        if hasattr(self, '_line_joints'):
            self._line_joints = ap.String(ap.LineJoints.MITER.value)
        self._line_dot_setting = None
        self._line_dash_setting = None
        self._line_round_dot_setting = None
        self._line_dash_dot_setting = None



"""Class implementation for the child-related interface.
"""

from typing import TYPE_CHECKING
from typing import Any
from typing import Dict
from typing import List
from typing import Union

from apysc._display.display_object import DisplayObject
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.array import Array
from apysc._type.boolean import Boolean
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface

if TYPE_CHECKING:
    from apysc._display.stage import Stage


class ChildInterface(RevertInterface):

    _children: Array[DisplayObject]
    _variable_name: str
    stage: 'Stage'

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='ChildInterface')
    def add_child(self, child: DisplayObject) -> None:
        """
        Add display object child to this instance.

        Parameters
        ----------
        child : DisplayObject
            Child instance to add.

        References
        ----------
        - add_child and remove_child interfaces document
            - https://simon-ritchie.github.io/apysc/add_child_and_remove_child.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite_1: ap.Sprite = ap.Sprite()
        >>> sprite_1.graphics.begin_fill(color='#0af')
        >>> rectangle: ap.Rectangle = sprite_1.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> sprite_2: ap.Sprite = ap.Sprite()
        >>> sprite_2.add_child(rectangle)
        """
        from apysc._validation import display_validation
        self._initialize_children_if_not_initialized()
        display_validation.validate_display_object(display_object=child)
        if child.parent is not None:
            child.remove_from_parent()
        self._children.append(child)
        child.parent = self
        append_expression_of_add_child(child=child)

    def _initialize_children_if_not_initialized(self) -> None:
        """
        Initialize _children attribute if this interface does not
        initialize it yet.
        """
        if hasattr(self, '_children'):
            return
        self._children = Array([])

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='ChildInterface')
    def remove_child(self, child: DisplayObject) -> None:
        """
        Remove display object child from this instance.

        Parameters
        ----------
        child : DisplayObject
            Child instance to remove.

        References
        ----------
        - add_child and remove_child interfaces document
            - https://simon-ritchie.github.io/apysc/add_child_and_remove_child.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> sprite.graphics.remove_child(rectangle)
        >>> print(rectangle.parent)
        None
        """
        self._initialize_children_if_not_initialized()
        append_expression_of_remove_child(child=child)
        for child_ in self._children.value:  # type: ignore
            if child_ != child:
                continue
            self._children.remove(child)
            child.parent = None
            return

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='ChildInterface')
    def contains(self, child: DisplayObject) -> Boolean:
        """
        Get a boolean whether this instance contains a specified child.

        Parameters
        ----------
        child : DisplayObject
            Child instance to check.

        Returns
        -------
        result : Boolean
            If this instance contains a specified child, this
            method returns True.

        References
        ----------
        - contains interface document
            - https://simon-ritchie.github.io/apysc/contains.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> sprite.graphics.contains(rectangle)
        Boolean(True)

        >>> rectangle.remove_from_parent()
        >>> sprite.graphics.contains(rectangle)
        Boolean(False)
        """
        import apysc as ap
        self._initialize_children_if_not_initialized()
        index: ap.Int = self._children.index_of(value=child)
        if index == -1:
            result: ap.Boolean = ap.Boolean(False)
        else:
            result = ap.Boolean(True)
        self._append_contains_expression(result=result, child=child)
        return result

    def _append_contains_expression(
            self, *, result: Boolean, child: DisplayObject) -> None:
        """
        Append contains method expression.

        Parameters
        ----------
        result : Boolean
            Result boolean value.
        child : DisplayObject
            Child instance to check.
        """
        import apysc as ap
        expression: str = (
            f'{result.variable_name} = '
            f'{self._variable_name}.has({child.variable_name});'
        )
        ap.append_js_expression(expression=expression)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='ChildInterface')
    def num_children(self) -> Int:
        """
        Get a current children's number.

        Returns
        -------
        num_children : int
            Current children number.

        References
        ----------
        - num_children interface document
            - https://simon-ritchie.github.io/apysc/num_children.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> rectangle_1: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle_2: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=150, y=50, width=50, height=50)
        >>> sprite.graphics.num_children
        Int(2)
        """
        import apysc as ap
        self._initialize_children_if_not_initialized()
        num_children: ap.Int = ap.Int(value=self._children.length)
        self._append_num_children_expression(num_children=num_children)
        return num_children

    def _append_num_children_expression(
            self, *, num_children: Int) -> None:
        """
        Append num_children method expression.

        Parameters
        ----------
        num_children : Int
            Current children number.
        """
        import apysc as ap
        expression: str = (
            f'{num_children.variable_name} = '
            f'{self._variable_name}.children().length;'
        )
        ap.append_js_expression(expression=expression)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='ChildInterface')
    def get_child_at(self, index: Union[int, Int]) -> DisplayObject:
        """
        Get child at a specified index.

        Parameters
        ----------
        index : int or Int
            Child's index (start from 0).

        Returns
        -------
        child : DisplayObject
            Target index child instance.

        References
        ----------
        - get_child_at interface document
            - https://simon-ritchie.github.io/apysc/get_child_at.html

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.begin_fill(color='#0af', alpha=0.5)
        >>> rectangle_1: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=50, y=50, width=50, height=50)
        >>> rectangle_2: ap.Rectangle = sprite.graphics.draw_rect(
        ...     x=150, y=50, width=50, height=50)
        >>> child_at_index_1: ap.DisplayObject = (
        ...     sprite.graphics.get_child_at(1))
        >>> child_at_index_1 == rectangle_2
        True
        """
        from apysc._expression import expression_variables_util
        from apysc._expression import var_names
        self._initialize_children_if_not_initialized()
        if self.num_children > index:
            child: DisplayObject = self._children[index]
        else:
            variable_name: str = expression_variables_util.\
                get_next_variable_name(type_name=var_names.DISPLAY_OBJECT)
            child = DisplayObject(variable_name=variable_name)
        self._append_get_child_at_expression(child=child, index=index)
        return child

    def _append_get_child_at_expression(
            self, *, child: DisplayObject, index: Union[int, Int]) -> None:
        """
        Append get_child_at method expression.

        Parameters
        ----------
        child : DisplayObject
            Target index child instance.
        index : int or Int
            Child's index (start from 0).
        """
        import apysc as ap
        from apysc._type import value_util
        index_str: str = value_util.get_value_str_for_expression(value=index)
        expression: str = (
            f'var {child.variable_name} = '
            f'{self._variable_name}.children()'
            f'[{index_str}];'
        )
        ap.append_js_expression(expression=expression)

    _children_snapshots: Dict[str, List[Any]]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make values' snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_children_if_not_initialized()
        if self._snapshot_exists(snapshot_name=snapshot_name):
            return

        for child in self._children.value:  # type: ignore
            if not isinstance(child, RevertInterface):
                continue
            child._run_all_make_snapshot_methods(snapshot_name=snapshot_name)

        self._set_single_snapshot_val_to_dict(
            dict_name='_children_snapshots',
            value=[*self._children._value], snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert values if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._children._value = [*self._children_snapshots[snapshot_name]]

        for child in self._children.value:  # type: ignore
            if not isinstance(child, RevertInterface):
                continue
            child._run_all_revert_methods(snapshot_name=snapshot_name)


@add_debug_info_setting(  # type: ignore[misc]
    module_name=__name__)
def append_expression_of_add_child(*, child: DisplayObject) -> None:
    """
    Append expression of add_child (add).

    Parameters
    ----------
    child : DisplayObject
        Child object to add.
    """
    import apysc as ap
    parent_name: str = child.parent.variable_name  # type: ignore
    child_name: str = child.variable_name
    expression: str = (
        f'{parent_name}.add({child_name});'
    )
    ap.append_js_expression(expression=expression)


@add_debug_info_setting(  # type: ignore[misc]
    module_name=__name__)
def append_expression_of_remove_child(*, child: DisplayObject) -> None:
    """
    Append expression of remove_child (removeElement).

    Parameters
    ----------
    child : DisplayObject
        Child object to remove.
    """
    import apysc as ap
    from apysc._expression import expression_variables_util
    from apysc._expression import var_names
    parent_name: str = expression_variables_util.get_next_variable_name(
        type_name=var_names.PARENT)
    child_name: str = child.variable_name
    expression: str = (
        f'var {parent_name} = {child_name}.parent();'
        f'\nif ({parent_name}) {{'
        f'\n  {parent_name}.removeElement({child_name});'
        '\n}'
    )
    ap.append_js_expression(expression=expression)



"""Event validation implementation.

Mainly following interfaces are defined.

- validate_event
    Validate specified instance is Event.
"""

from typing import Any

from apysc._event.event import Event
from apysc._event.mouse_event_type import MouseEventType


def validate_event(e: Any) -> Event:
    """
    Validate whether a specified instance is the
    `Event` class or not.

    Parameters
    ----------
    e : Event
        Event instance to check.

    Raises
    ------
    ValueError
        If a specified instance is not the `Event`
        class instance.

    Returns
    -------
    e : Event
        Event instance.
    """
    if isinstance(e, Event):
        return e
    raise ValueError(
        f'Specified instance is not Event type: {type(e)}')


def validate_event_type(mouse_event_type: Any) -> MouseEventType:
    """
    Validate whether specified value is MouseEventType one or not.

    Parameters
    ----------
    mouse_event_type : MouseEventType
        EventTMouseEventTypeype value to check.

    Returns
    -------
    mouse_event_type : MouseEventType
        MouseEventType value.
    """
    if isinstance(mouse_event_type, MouseEventType):
        return mouse_event_type
    raise ValueError(
        f'Specified value is not a MouseEventType: {type(mouse_event_type)}')



"""This module is for the translation mapping data of the
following document:

Document file: mouse_event_abstract.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# MouseEvent interfaces abstract':
    '',

    'This page explains the MouseEvent interfaces abstract.':
    '',

    '## What apysc can do in its interfaces':
    '',

    '- You can set the `MouseEvent` handlers, such as the click, mouse down, mouse over, and so on, to each graphic instance.\n- You can pass the optional arguments to the handler.':  # noqa
    '',

    '## Example of the click event':
    '',

    'To bind MouseEvent, defining the handler function (or method) would be necessary (e.g., `on_click`).\n\nThese handlers can bind with the click interface.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\n\ndef on_click(e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : ap.MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this\n    with ap.If(rectangle.fill_color == \'#00aaff\'):\n        rectangle.fill_color = ap.String(\'#f0a\')\n        ap.Return()\n\n    with ap.If(rectangle.fill_color == \'#ff00aa\'):\n        rectangle.fill_color = ap.String(\'#0af\')\n        ap.Return()\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.click(on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'mouse_event_abstract_click/\')\n```':  # noqa
    '',

    '<iframe src="static/mouse_event_abstract_click/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## See also':
    '',

    'There are a lot of other mouse event binding interfaces, such as the mouse down, mouse over, and mouse move. For more details, please see the following:\n\n- [Basic mouse event interfaces](mouse_event_basic.md)\n- [Click interface](click.md)\n- [Double click interface](dblclick.md)\n- [Mousedown and mouseup interfaces](mousedown_and_mouseup.md)\n- [Mouseover and mouseout interfaces](mouseover_and_mouseout.md)\n- [Mousemove interface](mousemove.md)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_draw_round_rect.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Graphics draw_round_rect interface':
    '',

    'This page explains the `Graphics` class `draw_round_rect` method interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    '`draw_round_rect` interface draws vector rounded rectangle graphics.':
    '',

    '## Basic usage':
    '',

    '`draw_rect` interface has `x`, `y`, `width`, and `height` arguments. `x` and `y` are rectangle coordinates setting, and `width` and `height` will determine rectangle size.\n\nThis interface also has `ellipse_width` and `ellipse_height` arguments to set the round size to the rectangle corners.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=350,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\n# Set 10-pixel ellipse size and draw the rectangle.\nsprite.graphics.draw_round_rect(\n    x=50, y=50, width=50, height=50, ellipse_width=10, ellipse_height=10)\n\n# Set 20-pixel ellipse size and draw the rectangle.\nsprite.graphics.draw_round_rect(\n    x=150, y=50, width=50, height=50, ellipse_width=20, ellipse_height=20)\n\n# Set 5-pixel ellipse width and 20-pixel ellipse height and\n# draw the rectangle.\nsprite.graphics.draw_round_rect(\n    x=250, y=50, width=50, height=50, ellipse_width=5, ellipse_height=20)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_round_rect_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_round_rect_basic_usage/index.html" width="350" height="150"></iframe>':  # noqa
    '',

    '## Return value':
    '',

    '`draw_round_rect` interface will return the `Rectangle` instance, same as the `draw_rect` interface.\n\nThe `Rectangle` instance has the `ellipse_width` attribute and `ellipse_height` to change the rectangle round size.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\nrectangle: ap.Rectangle = sprite.graphics.draw_round_rect(\n    x=50, y=50, width=50, height=50, ellipse_width=10, ellipse_height=10)\n\n# You can update the ellipse_width and ellipse_height\n# attributes dynamically.\nrectangle.ellipse_width = ap.Int(20)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_round_rect_return_value/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_round_rect_return_value/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## draw_round_rect API':
    '',

    '<!-- Docstring: apysc._display.graphics.Graphics.draw_round_rect -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `draw_round_rect(self, x:Union[int, apysc._type.int.Int], y:Union[int, apysc._type.int.Int], width:Union[int, apysc._type.int.Int], height:Union[int, apysc._type.int.Int], ellipse_width:Union[int, apysc._type.int.Int], ellipse_height:Union[int, apysc._type.int.Int]) -> apysc._display.rectangle.Rectangle`<hr>\n\n**[Interface summary]** Draw a rounded rectangle vector graphics.<hr>\n\n**[Parameters]**\n\n- `x`: Int or int\n  - X-coordinate to start drawing.\n- `y`: Int or int\n  - Y-coordinate to start drawing.\n- `width`: Int or int\n  - Rectangle width.\n- `height`: Int or int\n  - Rectangle height.\n- `ellipse_width`: Int or int\n  - Ellipse width of the rectangle corner.\n- `ellipse_height`: Int or int\n  - Ellipse height of the rectangle corner.\n\n<hr>\n\n**[Returns]**\n\n- `rectangle`: Rectangle\n  - Created rectangle.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> round_rect: ap.Rectangle = sprite.graphics.draw_round_rect(\n...     x=50, y=50, width=50, height=50,\n...     ellipse_width=10, ellipse_height=15)\n>>> round_rect.ellipse_width\nInt(10)\n\n>>> round_rect.ellipse_height\nInt(15)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: add_child_and_remove_child.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# add_child and remove_child interfaces':
    '',

    'This page explains the container class, like the `Graphics`\\, `Sprite`\\, `Stage`) `add_child` and `remove_child` method interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `add_child` and `remove_child` add or remove a `DisplayObject` child instance from a container instance. The apysc does not display a removed `DisplayObject` instance.':  # noqa
    '',

    '## Automatic addition of the children':
    '',

    'The apysc appends each `DisplayObject` instance to a parent at the constructor. So, for example, it appends a `Sprite` instance to a parent stage. Similarly, it appends a `graphics` instance to a parent `Sprite` instance.\n\nIf you need to adjust a parent, it is necessary to call the `add_child` or `remove_child` interfaces manually (for instance, set a `Sprite` parent to the other `Sprite`).':  # noqa
    '',

    '## Basic usage of the remove_child interface':
    '',

    'The `remove_child` interface removes a child from a parent container instance. The apysc does not display a removed `DisplayObject` instance.\n\nFor example, the following code calls the `remove_child` interface in the click handler, so if you click the rectangle, it removes that rectangle.':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_sprite_click(\n        e: ap.MouseEvent[ap.Sprite], options: _RectOptions) -> None:\n    """\n    The handler that the sprite calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    sprite: ap.Sprite = e.this\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    sprite.remove_child(child=rectangle)\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\noptions: _RectOptions = {\'rectangle\': rectangle}\nsprite.click(on_sprite_click, options=options)\n\nap.save_overall_html(\n    dest_dir_path=\'sprite_basic_usage_of_remove_child/\')\n```':  # noqa
    '',

    '<iframe src="static/sprite_basic_usage_of_remove_child/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## The basic usage of the add_child interface':
    '',

    'The `add_child` interface adds a removed child again or adds a child to the other container instance.\n\nThe following code example removes the rectangle from the first `Sprite` container (be positioned to the left) when you click the rectangle. Also, that click event adds the rectangle to the second `Sprite` container (be positioned to the right).':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _SpriteAndRectOptions(TypedDict):\n    rectangle: ap.Rectangle\n    sprite: ap.Sprite\n\n\ndef on_sprite_click(\n        e: ap.MouseEvent[ap.Sprite],\n        options: _SpriteAndRectOptions) -> None:\n    """\n    The handler that the sprite calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    first_sprite: ap.Sprite = e.this\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    second_sprite: ap.Sprite = options[\'sprite\']\n    first_sprite.remove_child(child=rectangle)\n    second_sprite.add_child(child=rectangle)\n\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=250,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nfirst_sprite: ap.Sprite = ap.Sprite()\nfirst_sprite.graphics.begin_fill(color=\'#0af\')\nfirst_sprite.x = ap.Int(50)\nfirst_sprite.y = ap.Int(50)\nrectangle: ap.Rectangle = first_sprite.graphics.draw_rect(\n    x=0, y=0, width=50, height=50)\n\nsecond_sprite: ap.Sprite = ap.Sprite()\nsecond_sprite.x = ap.Int(150)\nsecond_sprite.y = ap.Int(50)\n\noptions: _SpriteAndRectOptions = {\n    \'rectangle\': rectangle, \'sprite\': second_sprite}\nfirst_sprite.click(on_sprite_click, options=options)\n\nap.save_overall_html(\n    dest_dir_path=\'sprite_basic_usage_of_add_child/\')\n```':  # noqa
    '',

    '<iframe src="static/sprite_basic_usage_of_add_child/index.html" width="250" height="150"></iframe>':  # noqa
    '',

    '## See also':
    '',

    '- [DisplayObject class parent interfaces](display_object_parent.md)\n- [Contains interface](contains.md)':  # noqa
    '',

    '## add_child API':
    '',

    '<!-- Docstring: apysc._display.child_interface.ChildInterface.add_child -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `add_child(self, child:apysc._display.display_object.DisplayObject) -> None`<hr>\n\n**[Interface summary]** Add display object child to this instance.<hr>\n\n**[Parameters]**\n\n- `child`: DisplayObject\n  - Child instance to add.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite_1: ap.Sprite = ap.Sprite()\n>>> sprite_1.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite_1.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> sprite_2: ap.Sprite = ap.Sprite()\n>>> sprite_2.add_child(rectangle)\n```':  # noqa
    '',

    '':
    '',

    '## remove_child API':
    '',

    '<!-- Docstring: apysc._display.child_interface.ChildInterface.remove_child -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `remove_child(self, child:apysc._display.display_object.DisplayObject) -> None`<hr>\n\n**[Interface summary]** Remove display object child from this instance.<hr>\n\n**[Parameters]**\n\n- `child`: DisplayObject\n  - Child instance to remove.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\', alpha=0.5)\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> sprite.graphics.remove_child(rectangle)\n>>> print(rectangle.parent)\nNone\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_line_alpha.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Graphics line_alpha interface':
    '',

    'This page explains the `Graphics` class `line_alpha` property interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    'The `line_alpha` property interface updates or get the instance\'s line alpha (opacity).':  # noqa
    '',

    '## Basic usage':
    '',

    'The getter or setter interface value becomes (or requires) the `Number` value (0.0 to 1.0).\n\nThe following example sets the 0.5 line alpha to the second rectangle and 0.25 to the third rectangle:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=350, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.line_style(color=\'#0af\', thickness=5)\n\nrectangle_1: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\n\nrectangle_2: ap.Rectangle = sprite.graphics.draw_rect(\n    x=150, y=50, width=50, height=50)\nrectangle_2.line_alpha = ap.Number(0.5)\n\nrectangle_3: ap.Rectangle = sprite.graphics.draw_rect(\n    x=250, y=50, width=50, height=50)\nrectangle_3.line_alpha = ap.Number(0.25)\n\nap.save_overall_html(\n    dest_dir_path=\'./graphics_line_alpha_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_line_alpha_basic_usage/index.html" width="350" height="150"></iframe>':  # noqa
    '',

    '## line_alpha property API':
    '',

    '<!-- Docstring: apysc._display.line_alpha_interface.LineAlphaInterface.line_alpha -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get this instance\'s line alpha (opacity).<hr>\n\n**[Returns]**\n\n- `line_alpha`: Number\n  - Current line alpha (opacity. 0.0 to 1.0).\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(color=\'#fff\', thickness=5, alpha=1.0)\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> rectangle.line_alpha = ap.Number(0.5)\n>>> rectangle.line_alpha\nNumber(0.5)\n```':  # noqa
    '',

}



"""Implementation of mouse wheel event interfaces.

Notes
-----
Not supported each SVG elements' mouse wheel event currently, only
supported document (overall screen) mouse wheel.
"""

from typing import Callable
from typing import Optional
from typing import TypeVar

from apysc._event.wheel_event import WheelEvent
from apysc._html.debug_mode import add_debug_info_setting

_O = TypeVar('_O')
_Handler = Callable[[WheelEvent, _O], None]


@add_debug_info_setting(  # type: ignore[misc]
    module_name=__name__)
def bind_wheel_event_to_document(
        handler: _Handler[_O],
        *,
        options: Optional[_O] = None) -> str:
    """
    Bind wheel event to document (overall window).

    Parameters
    ----------
    handler : _Handler
        Callable that handles wheel event.
    options : dict or None, default None
        Optional arguments dictionary to pass.

    Returns
    -------
    name : str
        Handler's name.

    References
    ----------
    - About the handler options’ type document
        - https://simon-ritchie.github.io/apysc/about_handler_options_type.html  # noqa
    """
    import apysc as ap
    from apysc._event.handler import HandlerData
    from apysc._event.handler import append_handler_expression
    from apysc._event.handler import get_handler_name
    from apysc._validation.handler_options_validation import \
        validate_options_type
    validate_options_type(options=options)
    name: str = get_handler_name(handler=handler, instance=ap.document)
    expression: str = (
        f'$({ap.document.variable_name}).on("mousewheel", {name});'
    )
    ap.append_js_expression(expression=expression)

    if options is None:
        options = {}  # type: ignore
    handler_data: HandlerData = {  # type: ignore
        'handler': handler,  # type: ignore
        'options': options,
    }
    e: ap.WheelEvent = ap.WheelEvent(this=ap.document)
    append_handler_expression(
        handler_data=handler_data, handler_name=name, e=e)
    return name


@add_debug_info_setting(  # type: ignore[misc]
    module_name=__name__)
def unbind_wheel_event_from_document(handler: _Handler[_O]) -> None:
    """
    Unbind a specified handler's wheel event from a document
    (overall window).

    Parameters
    ----------
    handler : _Handler
        Callable to unbind.
    """
    import apysc as ap
    from apysc._event.handler import get_handler_name
    name: str = get_handler_name(handler=handler, instance=ap.document)
    expression: str = (
        f'$({ap.document.variable_name}).off("mousewheel", {name});'
    )
    ap.append_js_expression(expression=expression)


@add_debug_info_setting(  # type: ignore[misc]
    module_name=__name__)
def unbind_wheel_event_all_from_document() -> None:
    """
    Unbind all wheels event from the document (overall window).
    """
    import apysc as ap
    expression: str = (
        f'$({ap.document.variable_name}).off("mousewheel");'
    )
    ap.append_js_expression(expression=expression)



"""This module is for the translation mapping data of the
following document:

Document file: graphics_draw_round_dotted_line.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Graphics draw_round_dotted_line interface':
    '',

    'This page explains the `Graphics` class `draw_round_dotted_line` method interface.':  # noqa
    '',

    '## What interface is this?':
    '',

    '`draw_round_dotted_line` interface draws the simple straight round dotted-line graphics. This interface ignores `dot_setting`\\, `dash_setting`\\, `round_dot_setting`\\, `dash_dot_setting`\\, and `cap` settings (this interface is using round cap setting so cap setting will also be ignored).':  # noqa
    '',

    '## Basic usage':
    '',

    '`draw_round_dotted_line` interface has basic coordinates arguments of `x_start`, `y_start`, `x_end` and `y_end`. That also has `round_size` and `space_size` arguments to determine the round style (line round size and the space size between each round).':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=250,\n    stage_height=130,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\n\n# Set 5-pixel round size and draw the line.\nsprite.graphics.line_style(color=\'#0af\')\nsprite.graphics.draw_round_dotted_line(\n    x_start=50, y_start=50, x_end=200, y_end=50,\n    round_size=5, space_size=5)\n\n# Set 10-pixel round size and draw the line.\nsprite.graphics.draw_round_dotted_line(\n    x_start=50, y_start=80, x_end=200, y_end=80,\n    round_size=10, space_size=5)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_round_dotted_line_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_round_dotted_line_basic_usage/index.html" width="250" height="130"></iframe>':  # noqa
    '',

    '## Notes':
    '',

    'Since this interface uses the round cap setting, the line length becomes longer by the size of the cap.\n\nIf you want to align the left line position with other lines, subtract half-round size from the `x_start` argument.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=270,\n    stage_height=130,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\n\n# Set 5-pixel round size and draw the line.\nsprite.graphics.line_style(color=\'#0af\')\nsprite.graphics.draw_round_dotted_line(\n    x_start=50, y_start=50, x_end=220, y_end=50,\n    round_size=10, space_size=5)\n\n# Set 45-pixel (50 - half-round size) to x_start argument\n# and draw the normal line.\nsprite.graphics.draw_line(\n    x_start=45, y_start=80, x_end=225, y_end=80)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_round_dotted_line_notes/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_round_dotted_line_notes/index.html" width="270" height="130"></iframe>':  # noqa
    '',

    '## draw_round_dotted_line API':
    '',

    '<!-- Docstring: apysc._display.graphics.Graphics.draw_round_dotted_line -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `draw_round_dotted_line(self, x_start:Union[int, apysc._type.int.Int], y_start:Union[int, apysc._type.int.Int], x_end:Union[int, apysc._type.int.Int], y_end:Union[int, apysc._type.int.Int], round_size:Union[int, apysc._type.int.Int], space_size:Union[int, apysc._type.int.Int]) -> \'_line.Line\'`<hr>\n\n**[Interface summary]** Draw a round-dotted line vector graphics.<hr>\n\n**[Parameters]**\n\n- `x_start`: Int or int\n  - Line start x-coordinate.\n- `y_start`: Int or int\n  - Line start y-coordinate.\n- `x_end`: Int or int\n  - Line end x-coordinate.\n- `y_end`: Int or int\n  - Line end y-coordinate.\n- `round_size`: Int or int\n  - Dot round size.\n- `space_size`: Int or int\n  - Blank space size between dots.\n\n<hr>\n\n**[Returns]**\n\n- `line`: Line\n  - Created line graphics instance.\n\n<hr>\n\n**[Notes]**\n\nThis interface ignores line settings, like the `LineDotSetting`, except `LineRoundDotSetting`.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(color=\'#fff\', thickness=5)\n>>> line: ap.Line = sprite.graphics.draw_round_dotted_line(\n...    x_start=50, y_start=50, x_end=150, y_end=50,\n...    round_size=6, space_size=3)\n>>> line.line_color\nString(\'#ffffff\')\n\n>>> line.line_round_dot_setting.round_size\nInt(6)\n\n>>> line.line_round_dot_setting.space_size\nInt(3)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_x.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# animation_x interface':
    '',

    'This page explains the `animation_x` method interface.':
    '',

    '## What interface is this?':
    '',

    'The `animation_x` method interface creates an `AnimationX` instance. You can animate x-coordinate with it.\n\nThis interface exists on a `DisplayObject` subclass instance, such as the `Sprite` or `Rectangle` class.':  # noqa
    '',

    '## Basic usage':
    '',

    'The following example sets the x-coordinate animation (from 50 to 100) with the `animation_x` method:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nEASING: ap.Easing = ap.Easing.EASE_OUT_QUINT\nDURATION: int = 1000\n\n\ndef on_animation_complete_1(\n        e: ap.AnimationEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=50, duration=DURATION, easing=EASING)\n    animation_x.animation_complete(on_animation_complete_2)\n    animation_x.start()\n\n\ndef on_animation_complete_2(\n        e: ap.AnimationEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : ap.AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    animation_x: ap.AnimationX = rectangle.animation_x(\n        x=100, duration=DURATION, easing=EASING)\n    animation_x.animation_complete(on_animation_complete_1)\n    animation_x.start()\n\n\nap.Stage(\n    stage_width=200, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#00aaff\')\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nanimation_x: ap.AnimationX = rectangle.animation_x(\n    x=100, duration=DURATION, easing=EASING)\nanimation_x.animation_complete(on_animation_complete_1)\nanimation_x.start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_x_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_x_basic_usage/index.html" width="200" height="150"></iframe>':  # noqa
    '',

    '## Notes for the Circle and Ellipse classes':
    '',

    'The `Circle` and `Ellipse` classes\' `animation_x` interface will return an `AnimationCx` (center-x) class instance, instead of a `AnimationX` instance, for internal implementation reason.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=200, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#00aaff\')\ncircle: ap.Circle = sprite.graphics.draw_circle(x=100, y=100, radius=50)\nanimation_cx: ap.AnimationCx = circle.animation_x(\n    x=100, duration=1000, easing=ap.Easing.EASE_OUT_QUINT)\n```':  # noqa
    '',

    '':
    '',

    '## animation_x API':
    '',

    '<!-- Docstring: apysc._animation.animation_x_interface.AnimationXInterface.animation_x -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `animation_x(self, x:Union[int, apysc._type.int.Int], *, duration:Union[int, apysc._type.int.Int]=3000, delay:Union[int, apysc._type.int.Int]=0, easing:apysc._animation.easing.Easing=<Easing.LINEAR: \'function(x) {return x;}\'>) -> apysc._animation.animation_x.AnimationX`<hr>\n\n**[Interface summary]** Set the x-coordinate animation setting.<hr>\n\n**[Parameters]**\n\n- `x`: Int or int\n  - Destination of the x-coordinate.\n- `duration`: Int or int, default 3000\n  - Milliseconds before an animation ends.\n- `delay`: Int or int, default 0\n  - Milliseconds before an animation starts.\n- `easing`: Easing, default Easing.LINEAR\n  - Easing setting.\n\n<hr>\n\n**[Returns]**\n\n- `animation_x`: AnimationX\n  - Created animation setting instance.\n\n<hr>\n\n**[Notes]**\n\nTo start this animation, you need to call the `start` method of the returned instance.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_x(\n...     x=100,\n...     duration=1500,\n...     easing=ap.Easing.EASE_OUT_QUINT,\n... ).start()\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [Animation interfaces duration setting document](https://simon-ritchie.github.io/apysc/animation_duration.html)\n- [Animation interfaces delay setting document](https://simon-ritchie.github.io/apysc/animation_delay.html)\n- [Each animation interface return value document](https://simon-ritchie.github.io/apysc/animation_return_value.html)\n- [Sequential animation setting document](https://simon-ritchie.github.io/apysc/sequential_animation.html)\n- [animation_parallel interface document](https://simon-ritchie.github.io/apysc/animation_parallel.html)\n- [Easing enum document](https://simon-ritchie.github.io/apysc/easing_enum.html)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: graphics_draw_line.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Graphics draw_line interface':
    '',

    'This page explains the `Graphics` class `draw_line` method interface.':
    '',

    '## What interface is this?':
    '',

    '`draw_line` interface will draw the simple straight line graphics. This interface will ignore `dot_setting`, `dash_setting`, `round_dot_setting`, and `dash_dot_setting`.':  # noqa
    '',

    '## Basic usage':
    '',

    '`draw_line` inteface has `x_start` (line x-start coordinate), `y_start` (line y-start coordinate), `x_end` (line x-end coordinate), and `y_end` (line y-end coordinate) arguments.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=100,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\n\nsprite.graphics.line_style(color=\'#0af\', thickness=5)\nsprite.graphics.draw_line(x_start=50, y_start=50, x_end=150, y_end=50)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_line_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_line_basic_usage/index.html" width="200" height=100></iframe>':  # noqa
    '',

    '## Ignored line style settings':
    '',

    'This interface will ignore `dot_setting`, `dash_setting`, `round_dot_setting`, and `dash_dot_setting` for simplicity. If you need to draw these styled lines, then use `draw_dotted_line`, `draw_dashed_line`, `draw_round_dotted_line`, or `draw_dash_dotted_line` interfaces instead of the `draw_line` interface.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=100,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\n\n# dot_setting will be ignored, and the result line will not be dotted.\nsprite.graphics.line_style(\n    color=\'#0af\', thickness=5,\n    dot_setting=ap.LineDotSetting(dot_size=5))\nsprite.graphics.draw_line(x_start=50, y_start=50, x_end=150, y_end=50)\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_line_ignored_dot_setting/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_line_ignored_dot_setting/index.html" width="200" height=100></iframe>':  # noqa
    '',

    '## Line instance':
    '',

    '`draw_line` interface returns the `Line` instance. You can update each setting or bind events to that instance. `draw_dotted_line`, `draw_dashed_line`, `draw_round_dotted_line`\n, and `draw_dash_dotted_line` will also return the same type instance.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=200,\n    stage_height=100,\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\n\nsprite.graphics.line_style(\n    color=\'#0af\', thickness=5)\nline: ap.Line = sprite.graphics.draw_line(\n    x_start=50, y_start=50, x_end=150, y_end=50)\n\n# Update the line color from cyan to magenta.\nline.line_color = ap.String(\'#f0a\')\n\nap.save_overall_html(\n    dest_dir_path=\'graphics_draw_line_line_instance/\')\n```':  # noqa
    '',

    '<iframe src="static/graphics_draw_line_line_instance/index.html" width="200" height=100></iframe>':  # noqa
    '',

    '## draw_line API':
    '',

    '<!-- Docstring: apysc._display.graphics.Graphics.draw_line -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `draw_line(self, x_start:Union[int, apysc._type.int.Int], y_start:Union[int, apysc._type.int.Int], x_end:Union[int, apysc._type.int.Int], y_end:Union[int, apysc._type.int.Int]) -> \'_line.Line\'`<hr>\n\n**[Interface summary]** Draw a normal line vector graphic.<hr>\n\n**[Parameters]**\n\n- `x_start`: Int or int\n  - Line start x-coordinate.\n- `y_start`: Int or int\n  - Line start y-coordinate.\n- `x_end`: Int or int\n  - Line end x-coordinate.\n- `y_end`: Int or int\n  - Line end y-coordinate.\n\n<hr>\n\n**[Returns]**\n\n- `line`: Line\n  - Created line graphics instance.\n\n<hr>\n\n**[Notes]**\n\n ・This interface ignores line settings, like the `LineDotSetting`, `LineDashSetting`.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.line_style(color=\'#fff\', thickness=5)\n>>> line: ap.Line = sprite.graphics.draw_line(\n...     x_start=50, y_start=50, x_end=150, y_end=50)\n>>> line.line_color\nString(\'#ffffff\')\n\n>>> line.line_thickness\nInt(5)\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: continue.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Continue':
    '',

    'This page explains the `Continue` class.\n\nBefore reading on, maybe it is helpful to read the following page (apysc uses the `Continue` class for the same reason):\n\n- [Why the apysc library doesn\'t use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '',

    '## What is the Continue class?':
    '',

    'The `with For` block uses the `Continue` class to skip a current loop iteration (in JavaScript). It behaves like the Python built-in `continue` keyword.':  # noqa
    '',

    '## Basic usage':
    '',

    'The `Continue` class can only be used in the `with For` (or other loop class) block, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    stage_width=250, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\n\narr: ap.Array = ap.Array(range(2))\ni: ap.Int\nwith ap.For(arr) as i:\n    condition: ap.Boolean = i == 0\n    with ap.If(condition):\n        sprite.graphics.begin_fill(color=\'#0af\')\n        sprite.graphics.draw_rect(x=50, y=50, width=50, height=50)\n        ap.Continue()\n\n    sprite.graphics.begin_fill(color=\'#f0a\')\n    sprite.graphics.draw_rect(x=150, y=50, width=50, height=50)\n\nap.save_overall_html(\n    dest_dir_path=\'continue_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/continue_basic_usage/index.html" width="250" height="150"></iframe>\n\nIf you use the `Continue` class in the out of the `with For` block, then an exception is raised:':  # noqa
    '',

    '```py\nimport apysc as ap\n\nap.Continue()\n```':
    '',

    '':
    '',

    '```\nException: The `Continue` class can be instantiated in the with loop statement, for example, after the `with ap.For(...):` statement.\n```':  # noqa
    '',

    '':
    '',

    '## Continue API':
    '',

    '<!-- Docstring: apysc._loop._continue.Continue.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self) -> None`<hr>\n\n**[Interface summary]** The loop continue expression class.<hr>\n\n**[Notes]**\n\nThis class can be instantiated in the with loop statement, for example, after the `with ap.For(...):` statement.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> arr: ap.Array = ap.Array(range(3))\n>>> with ap.For(arr) as i:\n...     with ap.If(i == 1):\n...         _ = ap.Continue()\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: string.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# String':
    '',

    'This page explains the `String` class.\n\nBefore reading on, maybe it is helpful to read the following page:\n\n- [Why the apysc library doesn\'t use the Python built-in data type](why_apysc_doesnt_use_python_builtin_data_type.md)':  # noqa
    '',

    '## What is the String class?':
    '',

    'The `String` class is the apysc string class. It can accept `str` or `String` values at the constructor, as follows:':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nstring_1: ap.String = ap.String(\'Hello\')\nassert string_1 == \'Hello\'\n\nstring_2: ap.String = ap.String(string_1)\nassert string_2 == \'Hello\'\n```':  # noqa
    '',

    '':
    '',

    '## String class interfaces':
    '',

    'For more details about the `String` class each interface, please see the following:\n\n- [String class comparison operations](string_comparison_operations.md)\n- [String class addition and multiplication operations](string_addition_and_multiplication.md)':  # noqa
    '',

    '## String class constructor API':
    '',

    '<!-- Docstring: apysc._type.string.String.__init__ -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `__init__(self, value:Union[str, _ForwardRef(\'String\')]) -> None`<hr>\n\n**[Interface summary]** String class for apysc library.<hr>\n\n**[Parameters]**\n\n- `value`: String or str\n  - Initial string value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> string: ap.String = ap.String(\'Hello\')\n>>> string\nString(\'Hello\')\n\n>>> string += \' World!\'\n>>> string\nString(\'Hello World!\')\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [String class comparison operations document](https://simon-ritchie.github.io/apysc/string_comparison_operations.html)\n- [String class addition and multiplication operations document](https://simon-ritchie.github.io/apysc/string_addition_and_multiplication.html)':  # noqa
    '',

    '## value property API':
    '',

    '<!-- Docstring: apysc._type.string.String.value -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface summary]** Get a current string value.<hr>\n\n**[Returns]**\n\n- `value`: str\n  - Current string value.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> string: ap.String = ap.String(\'Hello\')\n>>> string.value = \'World!\'\n>>> string.value\n\'World!\'\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [apysc fundamental data classes value interface](https://simon-ritchie.github.io/apysc/fundamental_data_classes_value_interface.html)':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_scale_x_and_y_from_point.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# animation_scale_x_from_point and animation_scale_y_from_point interfaces':  # noqa
    '',

    'This page explains the `animation_scale_x_from_point` and `animation_scale_y_from_point` method interfaces.':  # noqa
    '',

    '## What interfaces are these?':
    '',

    'The `animation_scale_x_from_point` method interface will create an `ap.AnimationScaleXFromPoint` instance. You can animate the x-directional scale with it from a specified x-coordinate.\n\nSimilarly, the `animation_scale_y_from_point` method interface creates an `ap.AnimationScaleYFromPoint` instance. You can animate the y-directional scale with it from a specified y-coordinate.\n\nThese interfaces exist on a `GraphicsBase` subclass (that has the `scale_x_from_center` and `scale_y_from_center` interfaces), such as the `Rectangle` or `Circle`.':  # noqa
    '',

    '## Basic usage':
    '',

    'The following example sets the x-directional scale (left-side rectangle) and y-directional scale (right-side rectangle) animation (from 1.0 to 2.0) with the `animation_scale_x_from_point` and `animation_scale_y_from_point` methods.\n\nThese settings set the left-side rectangle scales from the left end (x=50) and the right-side rectangle scales from the bottom end (y=100).':  # noqa
    '',

    '```py\n# runnable\nfrom enum import Enum\n\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\nDURATION: int = 1000\nEASING: ap.Easing = ap.Easing.EASE_OUT_QUINT\nLEFT_RECTANGLE_X: int = 50\nRIGHT_RECTANGLE_Y: int = 100\nSCALE_1: float = 1.0\nSCALE_2: float = 2.0\n\n\nclass Direction(Enum):\n    X = 1\n    Y = 2\n\n\nclass Options(TypedDict):\n    direction: Direction\n\n\ndef on_animation_complete_1(\n        e: ap.AnimationEvent[ap.Rectangle], options: Options) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    if options[\'direction\'] == Direction.X:\n        rectangle.animation_scale_x_from_point(\n            scale_x_from_point=SCALE_1,\n            x=LEFT_RECTANGLE_X,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_2,\n            options=options,\n        ).start()\n    elif options[\'direction\'] == Direction.Y:\n        rectangle.animation_scale_y_from_point(\n            scale_y_from_point=SCALE_1,\n            y=RIGHT_RECTANGLE_Y,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_2,\n            options=options,\n        ).start()\n\n\ndef on_animation_complete_2(\n        e: ap.AnimationEvent[ap.Rectangle], options: Options) -> None:\n    """\n    The handler that the animation calls when its end.\n\n    Parameters\n    ----------\n    e : AnimationEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = e.this.target\n    if options[\'direction\'] == Direction.X:\n        rectangle.animation_scale_x_from_point(\n            scale_x_from_point=SCALE_2,\n            x=LEFT_RECTANGLE_X,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_1,\n            options=options,\n        ).start()\n    elif options[\'direction\'] == Direction.Y:\n        rectangle.animation_scale_y_from_point(\n            scale_y_from_point=SCALE_2,\n            y=RIGHT_RECTANGLE_Y,\n            duration=DURATION,\n            easing=EASING,\n        ).animation_complete(\n            on_animation_complete_1,\n            options=options,\n        ).start()\n\n\nap.Stage(\n    stage_width=250, stage_height=150,\n    background_color=\'#333\', stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\nleft_rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nright_rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=150, y=50, width=50, height=50)\n\noptions: Options = {\'direction\': Direction.X}\nleft_rectangle.animation_scale_x_from_point(\n    scale_x_from_point=SCALE_2,\n    x=LEFT_RECTANGLE_X,\n    duration=DURATION,\n    easing=EASING,\n).animation_complete(\n    on_animation_complete_1,\n    options=options,\n).start()\n\noptions = {\'direction\': Direction.Y}\nright_rectangle.animation_scale_y_from_point(\n    scale_y_from_point=SCALE_2,\n    y=RIGHT_RECTANGLE_Y,\n    duration=DURATION,\n    easing=EASING,\n).animation_complete(\n    on_animation_complete_1,\n    options=options,\n).start()\n\nap.save_overall_html(\n    dest_dir_path=\'./animation_scale_x_and_y_from_point_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_scale_x_and_y_from_point_basic_usage/index.html" width="250" height="150"></iframe>':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: animation_reverse.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# animation_reverse interface':
    '',

    'This page explains the `animation_reverse` method interface.':
    '',

    '## What interface is this?':
    '',

    'The `animation_reverse` interface reverses the running animations.\n\nThis interface exists in the instances that have the animation interfaces (such as the `animation_x`\\, `animation_move`).':  # noqa
    '',

    '## Basic usage':
    '',

    'The following example sets the x-coordinate animation and starts the 1-second interval timer to reverse animation with the `animation_reverse` interface.':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_timer_1(e: ap.TimerEvent, options: _RectOptions) -> None:\n    """\n    The event handler that timer calls after the 3 seconds.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.Timer(on_timer_2, delay=1000, options=options).start()\n\n\ndef on_timer_2(e: ap.TimerEvent, options: _RectOptions) -> None:\n    """\n    The event handler that timer calls every second.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n    rectangle.animation_reverse()\n\n\nap.Stage(\n    stage_width=500, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.animation_x(x=400, duration=5000).start()\noptions: _RectOptions = {\'rectangle\': rectangle}\nap.Timer(\n    on_timer_1, delay=3000, repeat_count=1,\n    options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'animation_reverse_basic_usage/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_reverse_basic_usage/index.html" width="500" height="150"></iframe>':  # noqa
    '',

    '## Interface Notes':
    '',

    'This interface can only use during animation. If you use this at the end of the animation, nothing happens, as follows:':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _RectOptions(TypedDict):\n    rectangle: ap.Rectangle\n\n\ndef on_timer(e: ap.TimerEvent, options: _RectOptions) -> None:\n    """\n    The handler that the timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    rectangle: ap.Rectangle = options[\'rectangle\']\n\n    # Nothing happens since the animation has already been completed.\n    rectangle.animation_reverse()\n\n\nap.Stage(\n    stage_width=500, stage_height=150, background_color=\'#333\',\n    stage_elem_id=\'stage\')\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.animation_x(x=400, duration=1000).start()\n\noptions: _RectOptions = {\'rectangle\': rectangle}\nap.Timer(on_timer, delay=1500, repeat_count=1, options=options).start()\n\nap.save_overall_html(\n    dest_dir_path=\'animation_reverse_notes/\')\n```':  # noqa
    '',

    '<iframe src="static/animation_reverse_notes/index.html" width="500" height="150"></iframe>':  # noqa
    '',

    '## animation_reverse API':
    '',

    '<!-- Docstring: apysc._animation.animation_reverse_interface.AnimationReverseInterface.animation_reverse -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `animation_reverse(self) -> None`<hr>\n\n**[Interface summary]** Reverse all running animations.<hr>\n\n**[Notes]**\n\nSuppose you call this interface multiple times and animations reach the beginning or end of the animation. In that case, this interface ignores the reverse instruction. This behavior means that the same interval\'s timer tick reverse setting does not work correctly (since the same interval setting reaches the animation start).<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> from typing_extensions import TypedDict\n>>> import apysc as ap\n>>> class RectOptions(TypedDict):\n...     rectangle: ap.Rectangle\n>>> def on_timer(\n...         e: ap.TimerEvent,\n...         options: RectOptions) -> None:\n...     rectangle: ap.Rectangle = options[\'rectangle\']\n...     rectangle.animation_reverse()\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.animation_x(\n...     x=100,\n...     duration=1500,\n...     easing=ap.Easing.EASE_OUT_QUINT,\n... ).start()\n>>> options: RectOptions = {\'rectangle\': rectangle}\n>>> ap.Timer(on_timer, delay=750, options=options).start()\n```':  # noqa
    '',

}



"""This module is for the translation mapping data of the
following document:

Document file: click.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# Click interface':
    '',

    'This page explains the `click` interface.':
    '',

    '## What interface is this?':
    '',

    'The `click` interface binds the click event to any `DisplayObject` instance (e.g., `Sprite`\\, `Rectangle`\\, and so on). These interfaces call the registered handler function if you mouse down on that instance and mouse up.\n\nConversely, the `unbind_click` interface unbinds the click event from the `DisplayObject` instance.':  # noqa
    '',

    '## See also':
    '',

    'The following page describes basic mouse event interfaces.\n\n- [Basic mouse event interfaces](mouse_event_basic.md)':  # noqa
    '',

    '## Basic usage of the click interface':
    '',

    'Each `DisplayObject` instance has the `click` method, and you can bind handlers by that.\n\nThe following example binds the click event handler to the rectangle.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\n\ndef on_click(\n        e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments.\n    """\n    rectangle: ap.Rectangle = e.this\n    rectangle.fill_color = ap.String(\'#f0a\')\n\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.click(handler=on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'click_basic_usage_of_the_click_interface/\')\n```':  # noqa
    '',

    'If you click the following rectangle, the rectangle color becomes the magenta color.\n\n<iframe src="static/click_basic_usage_of_the_click_interface/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## Basic usage of the unbind_click interface':
    '',

    'The `unbind_click` interface can remove the binded click event from a `DisplayObject` instance.\n\nThe following example removes the click event by the `unbind_click` method, and nothing happens if you click the rectangle.':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\n\ndef on_click(\n        e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments.\n    """\n    rectangle: ap.Rectangle = e.this\n    rectangle.fill_color = ap.String(\'#f0a\')\n\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.click(handler=on_click)\nrectangle.unbind_click(handler=on_click)\n\nap.save_overall_html(\n    dest_dir_path=\'click_basic_usage_of_the_unbind_click_interface/\')\n```':  # noqa
    '',

    '<iframe src="static/click_basic_usage_of_the_unbind_click_interface/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## Unbind all the click event handlers':
    '',

    '`unbind_click_all` interface can unbind all the click event handlers from the `DisplayObject` instance.\n\nThe following example removes all the click events by the `unbind_click_all` method (if you click the rectangle, nothing happens).':  # noqa
    '',

    '```py\n# runnable\nimport apysc as ap\n\nap.Stage(\n    background_color=\'#333\',\n    stage_width=150,\n    stage_height=150,\n    stage_elem_id=\'stage\')\n\nsprite: ap.Sprite = ap.Sprite()\nsprite.graphics.begin_fill(color=\'#0af\')\n\n\ndef on_click(\n        e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n    """\n    The handler that the rectangle calls when clicked.\n\n    Parameters\n    ----------\n    e : MouseEvent\n        Event instance.\n    options : dict\n        Optional arguments.\n    """\n    rectangle: ap.Rectangle = e.this\n    rectangle.fill_color = ap.String(\'#f0a\')\n\n\nrectangle: ap.Rectangle = sprite.graphics.draw_rect(\n    x=50, y=50, width=50, height=50)\nrectangle.click(handler=on_click)\nrectangle.unbind_click_all()\n\nap.save_overall_html(\n    dest_dir_path=\'click_unbind_all_the_click_event_handlers/\')\n```':  # noqa
    '',

    '<iframe src="static/click_unbind_all_the_click_event_handlers/index.html" width="150" height="150"></iframe>':  # noqa
    '',

    '## click API':
    '',

    '<!-- Docstring: apysc._event.click_interface.ClickInterface.click -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `click(self, handler:Callable[[apysc._event.mouse_event.MouseEvent, ~_O], NoneType], *, options:Union[~_O, NoneType]=None) -> str`<hr>\n\n**[Interface summary]** Add a click event listener setting.<hr>\n\n**[Parameters]**\n\n- `handler`: _Handler\n  - A callable would be called when clicking this instance.\n- `options`: dict or None, default None\n  - Optional arguments dictionary to be passed to a handler.\n\n<hr>\n\n**[Returns]**\n\n- `name`: str\n  - Handler\'s name.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> def on_click(\n...         e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n...     rectangle: ap.Rectangle = e.this\n...     rectangle.x += 10\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.click(on_click)\n```':  # noqa
    '',

    '<hr>\n\n**[References]**\n\n- [About the handler options’ type document](https://simon-ritchie.github.io/apysc/about_handler_options_type.html)':  # noqa
    '',

    '## unbind_click API':
    '',

    '<!-- Docstring: apysc._event.click_interface.ClickInterface.unbind_click -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `unbind_click(self, handler:Callable[[apysc._event.mouse_event.MouseEvent, ~_O], NoneType]) -> None`<hr>\n\n**[Interface summary]** Unbind specified handler\'s click event.<hr>\n\n**[Parameters]**\n\n- `handler`: _Handler\n  - Unbinding target Callable.\n\n<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> def on_click(\n...         e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n...     rectangle: ap.Rectangle = e.this\n...     rectangle.fill_color = ap.String(\'#f0a\')\n...     rectangle.unbind_click(on_click)\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.click(on_click)\n```':  # noqa
    '',

    '':
    '',

    '## unbind_click_all API':
    '',

    '<!-- Docstring: apysc._event.click_interface.ClickInterface.unbind_click_all -->\n\n<span class="inconspicuous-txt">Note: the document build script generates and updates this API document section automatically. Maybe this section is duplicated compared with previous sections.</span>\n\n**[Interface signature]** `unbind_click_all(self) -> None`<hr>\n\n**[Interface summary]** Unbind all click events.<hr>\n\n**[Examples]**':  # noqa
    '',

    '```py\n>>> import apysc as ap\n>>> def on_click(\n...         e: ap.MouseEvent[ap.Rectangle], options: dict) -> None:\n...     rectangle: ap.Rectangle = e.this\n...     rectangle.fill_color = ap.String(\'#f0a\')\n...     rectangle.unbind_click_all()\n>>> stage: ap.Stage = ap.Stage()\n>>> sprite: ap.Sprite = ap.Sprite()\n>>> sprite.graphics.begin_fill(color=\'#0af\')\n>>> rectangle: ap.Rectangle = sprite.graphics.draw_rect(\n...     x=50, y=50, width=50, height=50)\n>>> _ = rectangle.click(on_click)\n```':  # noqa
    '',

}



"""Interface class implementation for the y path data.
"""

from typing import Dict

from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.attr_linking_interface import AttrLinkingInterface
from apysc._type.int import Int
from apysc._type.revert_interface import RevertInterface


class PathYInterface(RevertInterface, AttrLinkingInterface):

    _y: Int

    def _initialize_y_if_not_initialized(self) -> None:
        """
        Initialize the _y attribute if this instance does not
        initialize it yet.
        """
        if hasattr(self, '_y'):
            return
        self._y = Int(0)

        self._append_y_linking_setting()

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='PathYInterface')
    def _append_y_linking_setting(self) -> None:
        """
        Append a y attribute linking settings.
        """
        self._append_applying_new_attr_val_exp(
            new_attr=self._y, attr_name='y')
        self._append_attr_to_linking_stack(
            attr=self._y, attr_name='y')

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='PathYInterface')
    def y(self) -> Int:
        """
        Get a y-coordinate of the destination point.

        Returns
        -------
        y : Int
            A y-coordinate of the destination point.

        Examples
        --------
        >>> import apysc as ap
        >>> line_to: ap.PathLineTo = ap.PathLineTo(x=50, y=50)
        >>> line_to.y = ap.Int(100)
        >>> line_to.y
        Int(100)
        """
        self._initialize_y_if_not_initialized()
        return self._y._copy()

    @y.setter
    def y(self, value: Int) -> None:
        """
        Set a y-coordinate of the destination point.

        Parameters
        ----------
        value : Int
            Y-coordinate of the destination point.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='y', args=[value], kwargs={},
                module_name=__name__,
                class_name=PathYInterface.__name__):
            self._initialize_y_if_not_initialized()
            self._y.value = value

            self._append_y_linking_setting()

    _y_snapshots: Dict[str, int]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make a value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_y_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_y_snapshots',
            value=int(self._y._value), snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert a value if a snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._initialize_y_if_not_initialized()
        self._y._value = self._y_snapshots[snapshot_name]



"""This module is for the translation mapping data of the
following document:

Document file: about_handler_options_type.md
Language: jp
"""

from typing import Dict

MAPPING: Dict[str, str] = {

    '# About the handler options\' type':
    '',

    'This page explains the event handler `options` argument\'s type.':
    '',

    '## The dictionary type is acceptable':
    '',

    'Each handler\'s `options` argument can accept a dictionary type value, like the following:':  # noqa
    '',

    '```py\n# runnable\nfrom typing import Dict\n\nimport apysc as ap\n\n\ndef on_timer(e: ap.TimerEvent, options: Dict[str, str]) -> None:\n    """\n    The handler that a timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.trace(options[\'msg\'])\n\n\ntimer: ap.Timer = ap.Timer(on_timer, delay=1000, options={\'msg\': \'Hello!\'})\ntimer.start()\n```':  # noqa
    '',

    '':
    '',

    '## About the TypedDict annotation':
    '',

    'Sometimes using the `TypedDict` type annotation instead of the `dict` type annotation is helpful and makes it easy to read the code. The apysc check a handler options annotation and the actual options value type when you use the `TypedDict`\\.':  # noqa
    '',

    '```py\n# runnable\nfrom typing_extensions import TypedDict\n\nimport apysc as ap\n\n\nclass _MsgOptions(TypedDict):\n    msg: str\n\n\ndef on_timer(e: ap.TimerEvent, options: _MsgOptions) -> None:\n    """\n    The handler that a timer calls.\n\n    Parameters\n    ----------\n    e : ap.TimerEvent\n        Event instance.\n    options : dict\n        Optional arguments dictionary.\n    """\n    ap.trace(options[\'msg\'])\n\n\noptions: _MsgOptions = {\'msg\': \'Hello!\'}\ntimer: ap.Timer = ap.Timer(on_timer, delay=1000, options=options)\ntimer.start()\n```':  # noqa
    '',

    'Notes: if you are using a Python 3.8 or later version, then importing the `TypedDict` from the `typing` package instead of the `typing_extensions` is available (e.g., `from typing import TypedDict`).':  # noqa
    '',

}



"""Else branch instruction implementation.
"""

from typing import Any
from typing import Dict
from typing import Optional

from apysc._branch.if_base import IfBase
from apysc._html.debug_mode import add_debug_info_setting


class Else(IfBase):
    """
    A class to append else branch instruction expression.

    Notes
    -----
    - You can only use this class immediately after the
        `If` or `Elif` statement.

    References
    ----------
    - Else document
        - https://simon-ritchie.github.io/apysc/else.html
    - Each branch instruction class's scope variables reverting setting
        - https://simon-ritchie.github.io/apysc/branch_instruction_variables_reverting_setting.html  # noqa

    Examples
    --------
    >>> import apysc as ap
    >>> int_val: ap.Int = ap.Int(10)
    >>> with ap.If(int_val >= 11):
    ...     ap.trace('Value is greater than equal 11.')
    >>> with ap.Else():
    ...     ap.trace('Value is less than 11.')
    """

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='Else')
    def __init__(
            self,
            *,
            locals_: Optional[Dict[str, Any]] = None,
            globals_: Optional[Dict[str, Any]] = None) -> None:
        """
        A class to append else branch instruction expression.

        Notes
        -----
        - You can only use this class immediately after the
            `If` or `Elif` statement.

        Parameters
        ----------
        locals_ : dict or None, default None
            Current scope's local variables. Set locals()
            value to this argument. If specified, this interface
            reverts all local scope VariableNameInterface
            variables (like Int, Sprite) at the end of an `Else`
            scope. This setting is useful when you don't want to
            update each variable by implementing the `Else` scope.
        globals_ : dict or None, default None
            Current scope's global variables. Set globals()
            value to this argument. This setting works
            the same way as the locals_ argument.

        References
        ----------
        - Else document
            - https://simon-ritchie.github.io/apysc/else.html
        - Each branch instruction class's scope variables reverting setting
            - https://simon-ritchie.github.io/apysc/branch_instruction_variables_reverting_setting.html  # noqa

        Examples
        --------
        >>> import apysc as ap
        >>> int_val: ap.Int = ap.Int(10)
        >>> with ap.If(int_val >= 11):
        ...     ap.trace('Value is greater than equal 11.')
        >>> with ap.Else():
        ...     ap.trace('Value is less than 11.')
        """
        super().__init__(
            condition=None,
            locals_=locals_,
            globals_=globals_)

    def _append_enter_expression(self) -> None:
        """
        Append else branch instruction start expression.

        Raises
        ------
        ValueError
            If the last scope is not If or Elif.
        """
        import apysc as ap
        if not self._last_scope_is_if_or_elif():
            raise ValueError(
                'Else interface can only use right after If or Elif '
                'interfaces.')
        expression: str = (
            'else {'
        )
        ap.append_js_expression(expression=expression)

    def _set_last_scope(self) -> None:
        """
        Set expression last scope value.
        """
        from apysc._expression import last_scope
        from apysc._expression.last_scope import LastScope
        last_scope.set_last_scope(value=LastScope.ELSE)



"""Class implementation for line joints interface.
"""

from typing import Dict
from typing import Union

from apysc._display.line_joints import LineJoints
from apysc._html.debug_mode import add_debug_info_setting
from apysc._type.revert_interface import RevertInterface
from apysc._type.string import String
from apysc._type.variable_name_interface import VariableNameInterface


class LineJointsInterface(VariableNameInterface, RevertInterface):

    _line_joints: String

    def _initialize_line_joints_if_not_initialized(self) -> None:
        """
        Initialize _line_joints attribute if this interface does
        not initialize it yet.
        """
        if hasattr(self, '_line_joints'):
            return
        self._line_joints = String(LineJoints.MITER.value)

    @property  # type: ignore[misc]
    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineJointsInterface')
    def line_joints(self) -> Union[String, LineJoints]:
        """
        Get this instance's line joints style setting.

        Returns
        -------
        line_joints : String
            Line joints style setting.

        Examples
        --------
        >>> import apysc as ap
        >>> stage: ap.Stage = ap.Stage()
        >>> sprite: ap.Sprite = ap.Sprite()
        >>> sprite.graphics.line_style(color='#fff', thickness=10)
        >>> _ = sprite.graphics.move_to(x=50, y=100)
        >>> _ = sprite.graphics.line_to(x=100, y=50)
        >>> line: ap.Polyline = sprite.graphics.line_to(x=150, y=100)
        >>> line.line_joints = ap.LineJoints.ROUND
        >>> line.line_joints
        String('round')
        """
        self._initialize_line_joints_if_not_initialized()
        return self._line_joints._copy()

    @line_joints.setter
    def line_joints(self, value: Union[String, LineJoints]) -> None:
        """
        Set line joints style setting.

        Parameters
        ----------
        value : String or LineJoints
            Line joints style setting to set.
        """
        from apysc._html.debug_mode import DebugInfo
        with DebugInfo(
                callable_='line_joints', args=[value], kwargs={},
                module_name=__name__,
                class_name=LineJointsInterface.__name__):
            self._update_line_joints_and_skip_appending_exp(value=value)
            self._append_line_joints_update_expression()

    def _update_line_joints_and_skip_appending_exp(
            self, *, value: Union[String, LineJoints]) -> None:
        """
        Update line joints and skip appending expression.

        Parameters
        ----------
        value : STring or LineJoints
            Line joints style setting to set.
        """
        from apysc._validation.display_validation import validate_line_joints
        if not isinstance(value, (String, LineJoints)):
            raise TypeError(
                'Not supported line_joints type specified: '
                f'{type(value)}'
                '\nAcceptable ones are: String or LineJoints.')
        validate_line_joints(joints=value)
        if isinstance(value, String):
            self._line_joints = value._copy()
        else:
            self._line_joints = String(value.value)

    @add_debug_info_setting(  # type: ignore[misc]
        module_name=__name__, class_name='LineJointsInterface')
    def _append_line_joints_update_expression(self) -> None:
        """
        Append line cap updating expression.
        """
        import apysc as ap
        from apysc._type import value_util
        joints_name: str = value_util.get_value_str_for_expression(
            value=self._line_joints)
        expression: str = (
            f'{self.variable_name}.attr'
            f'({{"stroke-linejoin": {joints_name}}});'
        )
        ap.append_js_expression(expression=expression)

    _line_joints_snapshots: Dict[str, str]

    def _make_snapshot(self, *, snapshot_name: str) -> None:
        """
        Make value's snapshot.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        self._initialize_line_joints_if_not_initialized()
        self._set_single_snapshot_val_to_dict(
            dict_name='_line_joints_snapshots',
            value=self._line_joints._value, snapshot_name=snapshot_name)

    def _revert(self, *, snapshot_name: str) -> None:
        """
        Revert value if snapshot exists.

        Parameters
        ----------
        snapshot_name : str
            Target snapshot name.
        """
        if not self._snapshot_exists(snapshot_name=snapshot_name):
            return
        self._line_joints._value = self._line_joints_snapshots[snapshot_name]
